#!/usr/bin/perl -w -T
#    smb_network.cgi - Web interface to your SMB network
#    Copyright (C) 2000  Gordon Messmer <yinyang@eburg.com>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# An SMB resource can be thought of as an extended file path, in this form:
#  //SERVER/SHARE/CWD/FILE
# This should be the easiest way to consider any SMB resource, and this
# CGI will try to keep as much of this path filled in as possible.



# Debugging will print the commands run before running them.
$debug = 0;

# Not using images will speed up the browser just ever so slightly
# but maybe you don't like them.
$use_icons = 1;

# Don't use MIME if you just want files to download, never using
# applications to handle the files.
$use_mime = 1;

# Allowing users to upload files is optional
$allow_uploads = 1;



# Some regex's that will be reused through this script
$WORKGROUP_RE = '((?:[\w-]+(?: (?=[\w-]))?)+)';
$NODE_RE      = '((?:[\w-]+(?: (?=[\w-]))?)+)';
$SHARE_RE     = '((?:[\w-]+(?:[ \'](?=[\w-]))?)+)';
$USER_RE = '([\w\s]*)';
$PASS_RE = '(.*)';
$FILE_RE = '([^/\\\\:*?"<>|]*)';
$CWD_RE  = "(\\\\?(?:${FILE_RE}\\\\?)*)";
$DOSATTR_RE = '[DAHSR]+';
$DIRATTR_RE = 'D[AHSR]*';
$ARCATTR_RE = '[D]*A[HSR]*';
$HDNATTR_RE = '[DA]*H[SR]*';
$SYSATTR_RE = '[DAH]*S[R]*';
$ROATTR_RE  = '[DAHS]*R';



# set up the PATH for taint mode
$ENV{'PATH'} = "/bin:/usr/bin:/usr/local/samba/bin";

# Set up a new CGI object to communicate with the user
use CGI;
$query = new CGI;

# Give all variables a blank initial state
$USER = $PASS = $WORKGROUP = $SERVER = $SHARE = $CWD = $FILE = "";

# Configure variables from input
if( defined($query->cookie('USERNAME')) ) {
    $USER = $query->cookie('USERNAME');
}
if( defined($query->cookie('PASSWORD')) ) {
    $PASS = $query->cookie('PASSWORD');
}
if( defined($query->param('WORKGROUP')) ) {
    $WORKGROUP = $query->param('WORKGROUP');
}
if( defined($query->param('SERVER')) ) {
    $SERVER = $query->param('SERVER');
}
if( defined($query->param('SHARE')) ) {
    $SHARE = $query->param('SHARE');
}
if( defined($query->param('CWD')) ) {
    $CWD = $query->param('CWD');
}
if( defined($query->param('FILE')) ) {
    $FILE = $query->param('FILE');
}



# Un-taint input so it can't interfere with commands run
sub my_untaintMe {
    my( $var, $rex ) = @_;
    if( $var =~ /$rex/ ) {
	return $1;
    } else {
	return "";
    }
}
$USER = &my_untaintMe( $USER, "^$USER_RE" );
$PASS = &my_untaintMe( $PASS, "^$PASS_RE" );;
$WORKGROUP = &my_untaintMe( $WORKGROUP, "^$WORKGROUP_RE" );
$SERVER = &my_untaintMe( $SERVER, "^$NODE_RE" );
$SHARE = &my_untaintMe( $SHARE, "^$SHARE_RE" );
$CWD  = &my_untaintMe( $CWD, "^$CWD_RE" );
$FILE = &my_untaintMe( $FILE, "$FILE_RE\$" );

# Create URL encoded copies of each field for use in links
$WORKGROUPURL = &encode( $WORKGROUP );
$SERVERURL = &encode( $SERVER );
$SHAREURL = &encode( $SHARE );
$CWDURL = &encode( $CWD );

# Further clean up input which will be used on the command line
$WORKGROUP =~ s/([\`\$])/\\$1/g;
$SERVER =~ s/([\`\$])/\\$1/g;
$SHARE =~ s/([\`\$])/\\$1/g;
$CWD  =~ s/([\`\$])/\\$1/g;
$FILE =~ s/([\`\$])/\\$1/g;

# Set up new cookies
@cookies = ();
if( defined($query->param('COOKIE')) ) {
    if( $query->param('COOKIE') eq 'Set Identity' ) {
	$name_cookie = $query->cookie(-name=>'USERNAME',
				      -value=>$query->param('USERNAME'));
	$pass_cookie = $query->cookie(-name=>'PASSWORD',
				      -value=>$query->param('PASSWORD'));

	splice(@cookies, $#cookies+1, 0, $name_cookie);
	splice(@cookies, $#cookies+1, 0, $pass_cookie);

	$USER = $query->param('USERNAME');
	$PASS = $query->param('PASSWORD');
    }
}



sub debug_print {
    my $msg = $_[0];
    if( $debug == 1 ) {
	print STDERR "$msg\n";
    }
}


sub encode {
    my $orig = $_[0];
    $orig =~ s/ /+/g;
    $orig =~ s/([^+\w])/sprintf("%%%x", ord($1))/eg;
    return $orig;
}


sub get_network_layout {
    $master_browser = "";

    # Prepare a nmblookup command that will get the master browser's IP
    $mb_lookup_command = "nmblookup -M -";
    if( $WORKGROUP ne "" ) {
	$mb_lookup_command .= " \"$WORKGROUP\"";
    }

    # Execute nmblookup and parse its output for the IP
    &debug_print( "$mb_lookup_command" );
    @mb_out = `$mb_lookup_command`;
    $master_browser_ip = "";
    foreach $line (@mb_out) {
	# The output of this line is funny.  It looks like whitespace, but
	# doesn't match \s
	if( $line =~ /^([0-9.]+).*__MSBROWSE__/ ) {
	    $master_browser_ip = $1;
	    last;
	}
    }
    if( $master_browser_ip eq "" ) {
	print "Unable to find the master browser for this domain!";
	print "  This is a fatal error.";
	goto CGI_EXIT;
    }

    # Continue to resolve the master browser's name
    &debug_print( "nmblookup -A $master_browser_ip" );
    @mb_name_out = `nmblookup -A $master_browser_ip`;
    foreach $line (@mb_name_out) {
	if( $line =~ /^\s*$NODE_RE\s*<00>\s*-\s*\w/ ) {
	    $master_browser = $1;
	    last;
	}
    }
    if( $master_browser eq "" ) {
	print "Unable to find the master browser for this domain!";
	print "  This is a fatal error.";
	goto CGI_EXIT;
    }

    # If no workgroup has been defined by the user, get the default from
    # the master browser
    if( $WORKGROUP eq "" ) {
	foreach $line (@mb_name_out) {
	    if( $line =~ /^\s*$WORKGROUP_RE\s*<00>\s*-\s*<GROUP>\s*\w/ ) {
		$WORKGROUP = $1;
		$WORKGROUPURL = &encode($WORKGROUP);
		last;
	    }
	}
    }
    if( $SERVER eq "" ) {
	$SERVER = $master_browser;
	$SERVERURL = &encode($SERVER);
    }
}


sub get_network_info {
    if( $WORKGROUP eq "" || $SERVER eq "" ) {
	&get_network_layout();
    }

    # Global lists for efficiency's sake
    @workgroups = ();
    @servers = ();
    @shares = ();

    # Get a list of servers and workgroups available
    if( $USER ne "" ) {
	$ENV{'USER'} = "$USER%$PASS";
    }

    $smbclient_command = "smbclient -N -L \"$SERVER\"";

    &debug_print( "$smbclient_command" );
    @layout = `$smbclient_command`;
    $layout_section = '';
    foreach $line (@layout) {
	if( $line =~ /ERRDOS/ ) {
	    next;
	}

	if( $line =~ /Sharename.*Comment/ ) {
	    $layout_section = "Shares";
	    next;
	}
	if( $line =~ /Server.*Comment/ ) {
	    $layout_section = "Servers";
	    next;
	}
	if( $line =~ /Workgroup.*Master/ ) {
	    $layout_section = "Workgroups";
	    next;
	}
	if( $line =~ /---------|^\s*$/ ) {
	    next;
	}

	if( $layout_section eq "Shares" ) {
	    splice(@shares, $#shares+1, 0, $line);
	    next;
	}
	if( $layout_section eq "Servers" ) {
	    splice(@servers, $#servers+1, 0, $line);
	    next;
	}
	if( $layout_section eq "Workgroups" ) {
	    splice(@workgroups, $#workgroups+1, 0, $line);
	    next;
	}
    }

    chomp @workgroups;
    chomp @servers;
    chomp @shares;
}


sub print_network_layout {
    &print_workgroups();
    &print_servers();
    &print_shares();
}


sub print_workgroups {
    if( ! defined(@workgroups) ) {
	&get_network_info();
    }

    print <<WGTABLE;
<TABLE width="80%" border="0" cellspacing="2" align="center">
<TR>
<TR align="center"><TD bgcolor="#DDDDDD"><B>WORKGROUPS</B></TD></TR>
WGTABLE
    foreach $group (@workgroups) {
	if( $group =~ /^\s*$WORKGROUP_RE\s.*/ ) {
	    print "<TR><TD bgcolor=\"#EEEEEE\">";
	    print "<IMG src=\"/smbicons/wg1.png\">" if( $use_icons == 1 );
	    print "<A href=\"$ENV{'SCRIPT_NAME'}?WORKGROUP=" . &encode($1) . "\">$1</A>";
	    print "</TD></TR>\n";
	}
    }
    print "</TABLE><BR><BR>\n";
}


sub print_servers() {
    if( ! defined(@servers) ) {
	&get_network_info();
    }

    print <<SVTABLE;
<TABLE width="80%" border="0" cellspacing="2" align="center">
<TR>
<TR align="center"><TD bgcolor="#DDDDDD"><B>SERVERS</B></TD></TR>
SVTABLE
    foreach $server (@servers) {
	if( $server =~ /^\s*$NODE_RE\s.*/ ) {
	    print "<TR><TD bgcolor=\"#EEEEEE\">";
	    print "<IMG src=\"/smbicons/comp1.png\">" if( $use_icons == 1 );
	    print "<A href=\"$ENV{'SCRIPT_NAME'}?ACTION=BROWSE_SERVER&WORKGROUP=$WORKGROUPURL&SERVER=" . &encode($1) . "\">$1</A>";
	    print "</TD></TR>\n";
	}
    }
    print "</TABLE><BR><BR>\n";
}


sub print_shares() {
    if( ! defined(@shares) ) {
	&get_network_info();
    }

    print <<SRTABLE;
<TABLE width="80%" border="0" cellspacing="2" align="center">
<TR>
<TR align="center"><TD bgcolor="#DDDDDD"><B>SHARES</B></TD></TR>
SRTABLE
    foreach $share (@shares) {
	# Only print the disk shares that don't end with '$'
	if( $share =~ /^\s*$SHARE_RE\s*Disk/ ) {
	    print "<TR><TD bgcolor=\"#EEEEEE\">";
	    print "<IMG src=\"/smbicons/dir_close_lnk.png\">" if( $use_icons == 1 );
	    print "<A href=\"$ENV{'SCRIPT_NAME'}?ACTION=BROWSE_SHARE&WORKGROUP=$WORKGROUPURL&SERVER=$SERVERURL&SHARE=" . &encode($1) . "\">$1</A>";
	    print "</TD></TR>\n";
	}
    }
    print "</TABLE><BR><BR>\n";
}


sub print_cwd_sort_routine() {
    $a_is_dir = $b_is_dir = 0;
    if( $a =~ / $DIRATTR_RE\s+(\d+)\s+(\w\w\w \w\w\w .. ..:..:.. \d+)$/ ){ $a_is_dir = 1; }
    if( $b =~ / $DIRATTR_RE\s+(\d+)\s+(\w\w\w \w\w\w .. ..:..:.. \d+)$/ ){ $b_is_dir = 1; }
    if( $a_is_dir and not $b_is_dir ){ return -1; }
    if( $b_is_dir and not $a_is_dir ){ return  1; }
    return $a cmp $b;
}


sub print_cwd () {
    # Set up the username and password in an environment variable, so we don't
    # have to pass it to smbclient on the command line.
    if( $USER ne "" ) {
	$ENV{'USER'} = "$USER%$PASS";
    }

    $smbclient_command = "smbclient \"//$SERVER/$SHARE\" -N -c \"ls \\\"$CWD\\\\*\\\"\"";

    &debug_print( "$smbclient_command" );
    @files = `$smbclient_command`;
    chomp( @files );
    @files = sort print_cwd_sort_routine @files;

    print <<FLTABLE;
<TABLE width="80%" border="0" cellspacing="2" align="center">
<TR>
<TR align="center"><TD bgcolor="#DDDDDD" colspan="3"><B>FILES</B></TD></TR>
<TR>
    <TD bgcolor="#DDDDDD">Name</TD>
    <TD bgcolor="#DDDDDD">Size</TD>
    <TD bgcolor="#DDDDDD">Last Modified</TD>
</TR>
FLTABLE

    if( $CWDURL =~ /(.*)%5c.+/ ) {
	$PARENTURL = $1;
	print "<TR><TD bgcolor=\"#EEEEEE\">";
	print "<IMG src=\"/smbicons/dir_up.png\">" if( $use_icons == 1 );
	print "<A href=\"$ENV{'SCRIPT_NAME'}?ACTION=BROWSE_SHARE&WORKGROUP=$WORKGROUPURL&SERVER=$SERVERURL&SHARE=$SHAREURL&CWD=$PARENTURL\">Parent Directory</A>";
	print "</TD><TD bgcolor=\"#EEEEEE\"></TD><TD bgcolor=\"#EEEEEE\"></TD></TR>\n";
    }

    foreach $file (@files) {
	if( $file =~ /ERRDOS/ ) {
	    print "Server connection has closed";
	    last;
	}
	if( $file =~ /ERRbadpw|ERRSRV/ ) {
	    print "<TR><TD>Permission denied - ";
	    print "Try setting a username and password and try again.";
	    print "</TD></TR>";
	    last;
	}
	if( $file =~ /^  \. |^  \.\. / ) {
	    next;
	}
	# Complex regex:
	# matching lines will start with two spaces (smbclient messages won't,
	# but the files listed will) continue through any number of characters,
	# terminating in a non-space character, followed by a decimal string
	# indicating the file size, followed by a standard date indicating
	# the file's last modification time.
	if( $file =~ /^  (.*?)($DOSATTR_RE)?\s+(\d+)\s+(\w\w\w \w\w\w .. ..:..:.. \d+)$/) {
	    $file = $1;
	    $attrs = $2;
	    $fsize = $3;
	    $fmod = $4;

	    if( $attrs ) { $file =~ s/\s+$//; }
	    else { $attrs = ""; }
	    $fURL = &encode($file);

	    print "<TR><TD bgcolor=\"#EEEEEE\">";
	    if( $attrs =~ /D/ ) {
		print "<IMG src=\"/smbicons/dir_close.png\">" if( $use_icons == 1 );
		print "<A href=\"$ENV{'SCRIPT_NAME'}?ACTION=BROWSE_SHARE&WORKGROUP=$WORKGROUPURL&SERVER=$SERVERURL&SHARE=$SHAREURL&CWD=$CWDURL%5c$fURL\">$file</A>";
	    }
	    else
	    {
		# Normal files will have the same A tag, but different images
		# FIXME: We need icons for the other flags: AHSR
		if( $use_icons == 1 )
		{
		    if( $attrs =~ /A/ ) {
			print "<IMG src=\"/smbicons/page.png\">";
		    } elsif( $attrs =~ /H/ ) {
			print "<IMG src=\"/smbicons/dotfile.png\">";
		    } elsif( $attrs =~ /S/ ) {
			print "<IMG src=\"/smbicons/page.png\">";
		    } elsif( $attrs =~ /R/ ) {
			print "<IMG src=\"/smbicons/page.png\">";
		    } else {
			print "<IMG src=\"/smbicons/page.png\">";
		    }
		}
		print "<A href=\"$ENV{'SCRIPT_NAME'}?ACTION=DOWNLOAD&WORKGROUP=$WORKGROUPURL&SERVER=$SERVERURL&SHARE=$SHAREURL&CWD=$CWDURL&FILE=$fURL\">$file</A>";
	    }
	    print "</TD>";
	    print "<TD bgcolor=\"#EEEEEE\">$fsize</TD>";
	    print "<TD bgcolor=\"#EEEEEE\">$fmod</TD></TR>\n";
	}
    }
    print <<EFLTABLE;
<TR align="center"><TD bgcolor="#EEEEEE" colspan="3">
    <FORM action="$ENV{'SCRIPT_NAME'}" method="POST">
    <INPUT type="submit" name="SUBMIT" value="Create New Directory: ">
    <INPUT type="text" size="10" name="FILE">
    <INPUT type="hidden" name="ACTION" value="MKDIR">
    <INPUT type="hidden" name="WORKGROUP" value="$WORKGROUP">
    <INPUT type="hidden" name="SERVER" value="$SERVER">
    <INPUT type="hidden" name="SHARE" value="$SHARE">
    <INPUT type="hidden" name="CWD" value="$CWD">
    </FORM>
</TD></TR>
EFLTABLE

    if( $allow_uploads == 1 ) {
	print <<EUPLOADS;
<TR align="center"><TD bgcolor="#EEEEEE" colspan="3">
    <FORM action="$ENV{'SCRIPT_NAME'}" enctype="multipart/form-data" method="POST">
    <INPUT type="submit" name="SUBMIT" value="Upload File: ">
    <INPUT type="file" size="10" name="FILE">
    <INPUT type="hidden" name="ACTION" value="UPLOAD">
    <INPUT type="hidden" name="WORKGROUP" value="$WORKGROUP">
    <INPUT type="hidden" name="SERVER" value="$SERVER">
    <INPUT type="hidden" name="SHARE" value="$SHARE">
    <INPUT type="hidden" name="CWD" value="$CWD">
    </FORM>
</TD></TR>
EUPLOADS
    }

    print "</TABLE><BR><BR>\n";
}


sub smb_mkdir () {
    # Set up the username and password in an environment variable, so we don't
    # have to pass it to smbclient on the command line.
    if( $USER ne "" ) {
	$ENV{'USER'} = "$USER%$PASS";
    }

    $smbclient_command = "smbclient \"//$SERVER/$SHARE\" -N -c \"mkdir \\\"$CWD\\\\$FILE\\\"\"";

    &debug_print( "$smbclient_command" );
    @output = `$smbclient_command`;
    foreach $line (@output) {
	if( $line =~ /ERRDOS/ ) {
	    print "Server connection has closed";
	    last;
	}
	if( $line =~ /ERRbadpw|ERRSRV/ ) {
	    print "Permission denied creating dir - Try setting a username and password and try again.";
	    last;
	}
    }
}


sub smb_upload () {
    # Write the file contents to a temp file
    $tmp_file = `mktemp /tmp/smbnet.XXXXXX`;
    $tmp_file = &my_untaintMe( $tmp_file, '(/tmp/smbnet\.\w*)' );
    open( TMPFILE, ">$tmp_file" );

    $fh = $query->upload('FILE');
    if( not defined( $fh ) ) { print "Error getting file handle from form.<BR>"; }
    while( read($fh, $buffer, 1024) ) {
	print TMPFILE $buffer;
    }
    close( TMPFILE );

    # Set up the username and password in an environment variable, so we don't
    # have to pass it to smbclient on the command line.
    if( defined($USER) ) {
	$ENV{'USER'} = "$USER%$PASS";
    }

    $smbclient_command = "smbclient \"//$SERVER/$SHARE\" -N -c \"put $tmp_file \\\"$CWD\\\\$FILE\\\"\"";

    &debug_print( "Uploading file: \n"
	. `ls -l $tmp_file`
	. "$smbclient_command" );
    @output = `$smbclient_command`;
    foreach $line (@output) {
	if( $line =~ /ERRDOS|Error|ERRSRV/ ) {
	    print "<BR><DIV align=\"center\"><B>Error uploading file to server</B></DIV><BR>";
	    last;
	}
    }

    unlink $tmp_file;
}


sub smb_download () {
    # Set up the username and password in an environment variable, so we don't
    # have to pass it to smbclient on the command line.
    if( $USER ne "" ) {
	$ENV{'USER'} = "$USER%$PASS";
    }

    $tmp_file = `mktemp /tmp/smbnet.XXXXXX`;
    $tmp_file = &my_untaintMe( $tmp_file, '(/tmp/smbnet\.\w*)' );

    $smbclient_command = "smbclient \"//$SERVER/$SHARE\" -N -c \"get \\\"$CWD\\\\$FILE\\\" $tmp_file\"";

    &debug_print( "$smbclient_command" );
    @output = `$smbclient_command`;
    foreach $line (@output) {
	if( $line =~ /ERRDOS|Error|ERRSRV/ ) {
	    print $query->header('-type' => 'text/plain');
	    print "Error downloading file from server";
	    return;
	}
    }

    if( $use_mime ) {
	# Find the MIME type using the 'file' command
	$fmimetype = `file -i '$tmp_file'`;
	chomp $fmimetype;
	$fmimetype =~ s/.*: (.*)/$1/;
	if( $fmimetype eq "data" || $fmimetype eq "" ) {
	    $fmimetype = 'application/octet-stream';
	}
    } else {
	$fmimetype = 'application/octet-stream';
    }

    # Get the file size so the browser can have a progress indicator.
    $fsize = (stat( $tmp_file ))[7];

    print $query->header('-type' => $fmimetype,
			 '-Content-Disposition' => "filename=\"$FILE\"",
			 '-Content-Length' => $fsize);
    print `cat $tmp_file ; rm -f $tmp_file`;
}



################################################################################
# Print out cookies that need to be set, and the HTML header
# When downloading files, we don't send back any HTML, so this has to be
# a function.   HTMLEND is always printed if we get that far.
sub html_head () {
    my $title = "";
    if( defined($_[0]) ) {
	$title = ":  $_[0]";
    }
    print $query->header('-cookie' => [@cookies]);
    print <<HTMLHEAD;
<HTML>
<HEAD><TITLE>SMB Network ${title}</TITLE></HEAD>
<BODY bgcolor="white">

<DIV align="center">
    [<A href="$ENV{'SCRIPT_NAME'}">Entire Network</A>]
</DIV>

HTMLHEAD
}
################################################################################

if( ! defined($query->param('ACTION')) ) {
    &html_head();
    &print_network_layout();
}
elsif( $query->param('ACTION') eq "BROWSE_SERVER" ) {
    &html_head( "\\\\${SERVER}\\" );
    &print_shares();
}
elsif( $query->param('ACTION') eq 'BROWSE_SHARE' ) {
    &html_head( "\\\\${SERVER}\\${SHARE}${CWD}" );
    &print_cwd();
}
elsif( $query->param('ACTION') eq 'MKDIR' ) {
    &html_head( "\\\\${SERVER}\\${SHARE}${CWD}" );
    &smb_mkdir() if( $allow_uploads );
    &print_cwd();
}
elsif( $query->param('ACTION') eq 'UPLOAD' ) {
    &html_head( "\\\\${SERVER}\\${SHARE}${CWD}" );
    &smb_upload() if( $allow_uploads );
    &print_cwd();
}
elsif( $query->param('ACTION') eq 'DOWNLOAD' ) {
    &smb_download();
    exit;
}
else {
    &html_head();
    &print_network_layout();
}


################################################################################
# Print out the HTML tail and exit

CGI_EXIT:

print <<HTMLTAIL;

<DIV align="center">

<FORM action="$ENV{'SCRIPT_NAME'}" method="POST">
Username: <INPUT type="text" size="10" name="USERNAME" value="$USER">
Password: <INPUT type="password" size="10" name="PASSWORD">
<INPUT type="submit" name="COOKIE" value="Set Identity">
</FORM>

</BODY>
</HTML>

HTMLTAIL
